from io import StringIO
from fractions import Fraction
from itertools import cycle
from operator import itemgetter
from pathlib import Path
import pickle
from random import choices

from cytoolz import assoc, groupby, valmap
import pytest

from bourbaki.application.config import load_config
from bourbaki.application.cli.main import OPTIONAL_ARG_TEMPLATE
from bourbaki.application.typed_io.utils import *
from bourbaki.application.typed_io.cli_repr_ import bool_cli_repr
from bourbaki.introspection.callables import is_classmethod, is_staticmethod

DIR = Path(__file__).parent.absolute()
sys.path.insert(0, str(DIR))
from cli import cli, MyCommandLineApp, FooEnum


def maybe(key):
    return OPTIONAL_ARG_TEMPLATE.format(key)


output_args = {
    "pretty": False,
    "literal": False,
    maybe("outfile"): text_path_repr,
}

fooenum_repr = "|".join(v.name for v in FooEnum)

today = datetime.date.today().isoformat()

CONFIG_FILE = str(DIR / "conf.yml")

CONFIG = {
    "__init__": {"a": 1, "b": [1, 2, 3], maybe("c"): date_repr},
    "print": {
        "ns": {**output_args},
        "tuple": {
            "kwarg": {
                "tup": [
                    type_spec(int),
                    type_spec(int) + " OR " + fraction_repr,
                    type_spec(float),
                ],
                **output_args,
            },
            "pos": {
                "tup": [
                    type_spec(int),
                    type_spec(int) + " OR " + fraction_repr,
                    type_spec(float),
                ],
                **output_args,
            },
            "union-kwarg": {
                "tup": [
                    [type_spec(int) + " OR " + fraction_repr, type_spec(str)],
                    "OR",
                    [type_spec(str), type_spec(bool)],
                ],
                **output_args,
            },
        },
        "enum": {maybe("foo"): fooenum_repr, **output_args},
        "bytes": {"b": [byte_repr, ellipsis_], **output_args},
        "flags": {
            "boolean1": bool_cli_repr,
            "boolean2": False,
            "flag1": False,
            "flag2": True,
        },
        "mapping": {
            "foo_to_date": {fooenum_repr: date_repr, ellipsis_: ellipsis_},
            **output_args,
        },
        "named-tuple": {
            "tup": {"foo": FooEnum.foo.name, "bar": ".", "baz": today},
            **output_args,
        },
        "numbers": {
            maybe("x"): decimal_repr,
            maybe("y"): fraction_repr,
            maybe("z"): complex_repr,
            **output_args,
        },
        "url": {"url": url_repr, **output_args},
        "uuid": {maybe("uuid"): uuid_repr, **output_args},
    },
    "args": {"args": [fooenum_repr, ellipsis_], **output_args},
    "args-and-kwargs": {
        "ips": [ipv6_repr, "..."],
        **output_args,
        "named_ips": {type_spec(str): ipv4_repr, ellipsis_: ellipsis_},
    },
    "types": {
        "number": classpath_type_repr + "<:Union[int,fractions.Fraction]",
        **output_args,
        "types": {
            type_spec(str): classpath_type_repr + "<:Mapping",
            ellipsis_: ellipsis_,
        },
    },
    "leading": {
        "list": {"l_1": [float_repr, ellipsis_], "i_2": int_repr, **output_args}
    },
    "cant": {
        "parse": {
            maybe("cant_parse_me"): [[type_spec(str), ellipsis_], ellipsis_],
            "can_parse_me": [type_spec(str), ellipsis_],
            **output_args,
        }
    },
    "get": {"attr": {"attr": type_spec(str), **output_args}},
}


def keys_recursive(dict_: dict):
    def inner(dict_, prefix):
        recurse = all(isinstance(v, dict) for v in dict_.values())
        if recurse:
            for k, v in dict_.items():
                yield from inner(v, (*prefix, k))
        else:
            yield prefix

    return list(t for t in inner(dict_, ()) if not t[0].startswith("__"))


def getrecursive(dict_, keys):
    if not any(keys):
        return dict_
    head_to_tails = valmap(
        lambda l: [t[1:] for t in l], groupby(itemgetter(0), filter(len, keys))
    )
    return {
        head: getrecursive(dict_[head], tails) for head, tails in head_to_tails.items()
    }


COMMANDS = keys_recursive(CONFIG)

RANDOM_COMMANDS = [set(choices(COMMANDS, k=3)) for _ in range(10)]

fmts = ["py", "yaml", "json"]


# currently can't run this because lookup of classes defined in cli.py fails
# @pytest.fixture(scope='module')
# def cli_pickle_reloaded():
#     f = io.BytesIO()
#     pickle.dump(cli, f)
#     f.seek(0)
#     return pickle.load(f)


def test_command_names():
    fns = (
        getattr(v, "__command_prefix__", (n.replace("_", "-"),))[0]
        for n, v in MyCommandLineApp.__dict__.items()
        if callable(v)
        and not n.startswith("_")
        and not is_classmethod(MyCommandLineApp, n)
        and not is_staticmethod(MyCommandLineApp, n)
    )
    assert set(cli.subcommands) == set(fns).union(["init"])


# TODO: more tests
@pytest.mark.parametrize(
    "args, result, config_format",
    [
        (
            ["--config", CONFIG_FILE, "print", "ns"],
            {"a": 1, "b": [1, 2, 3], "c": datetime.date.today()},
            None,
        ),
        (
            ["--config", CONFIG_FILE, "print", "tuple", "kwarg"],
            (12345678, Fraction(1234, 5678), 1234.5678),
            None,
        ),
        (
            ["print", "tuple", "kwarg", "--tup", "123", "456", "7.89"],
            (123, 456, 7.89),
            None,
        ),
        *[(["init", "config", "--format", fmt], CONFIG, "." + fmt) for fmt in fmts],
        *[
            (
                [
                    "init",
                    "config",
                    "--format",
                    fmt,
                    "--only-commands",
                    *map(" ".join, cmds),
                ],
                assoc(getrecursive(CONFIG, cmds), "__init__", CONFIG["__init__"]),
                "." + fmt,
            )
            for cmds, fmt in zip(RANDOM_COMMANDS, cycle(fmts))
        ],
    ],
)
def test_cli_result(args, result, config_format, capsys):
    cli_result = cli.run(args)
    assert cli_result == result
    if config_format:
        out = capsys.readouterr().out
        parsed = load_config(StringIO(out), ext=config_format)
        assert parsed == result


@pytest.mark.parametrize("args,exc", [
    (['--install-bash-completion'], SystemExit),
    (['--version'], SystemExit),
    (["--info"], SystemExit),
    (['--execute', 'print', 'ns'], None),
])
def test_cli_special_actions(args, exc):
    if exc is None:
        cli.run(args)
    else:
        with pytest.raises(exc):
            cli.run(args)


def test_cli_parses_env(monkeypatch):
    monkeypatch.setenv('CLI_ARG_B', '"1/2" 34 \'56\'')
    ns = cli.run(['print', 'ns'])
    assert ns['b'] == [Fraction(1, 2), 34, 56]
    monkeypatch.delenv('CLI_ARG_B')



#!/usr/bin/env python
# coding:utf-8
import logging
import os
import pickle
from itertools import chain
from collections import Counter
from warnings import warn
import pytest
from bourbaki.application.logging import Logged, InstanceLoggerNamingConvention, configure_debug_logging
from bourbaki.application.logging.defaults import (
    DEFAULT_LOG_MSG_FMT,
    DEFAULT_LOG_DATE_FMT,
)
from bourbaki.application.logging.analysis import log_line_regex, log_file_to_df
from bourbaki.application.logging import CountingLogger, configure_default_logging

class _TestLogger(CountingLogger):
    def disable(self):
        self.disabled = True

    def _log(self, level, msg, *a, **kw):
        msg = " ".join((nextcounter(), msg or ""))
        super()._log(level, msg, *a, **kw)


logging.setLoggerClass(_TestLogger)

configure_default_logging(
    console_level=0,
    disable_existing_loggers=True,
)
logging.root.getChild('matplotlib').disable()
logging.root.getChild('pandas').disable()


global TestLoggedClass
MSG_COUNTER = 0
MSG_TYPES = ("debug", "info", "warn", "warning", "error", "critical")


def nextcounter():
    global MSG_COUNTER
    MSG_COUNTER += 1
    return str(MSG_COUNTER).zfill(4)


def get_test_instance(log_name: str):
    return TestLoggedClass("a", "few", "args", log_name=log_name)


@pytest.fixture(scope="module")
def this_logfile():
    path = "/tmp/test.log"
    yield path
    os.remove(path)


@pytest.fixture(scope="module")
def log_pickle_file():
    pickle_file = "/tmp/test_df.pkl"
    yield pickle_file
    os.remove(pickle_file)


@pytest.fixture
def test_pickle_path():
    path = "/tmp/test_instance.pkl"
    yield path
    os.remove(path)


@pytest.fixture
def logger():
    logger = logging.getLogger(__name__)
    return logger


def test_setup(this_logfile):
    configure_debug_logging(
        filename=this_logfile, dated_logfiles=False, disable_existing_loggers=True
    )

    global TestLoggedClass, _TestLogger

    class TestLoggedClass(Logged):
        __log_level__ = logging.DEBUG
        __verbose__ = True
        __instance_naming__ = InstanceLoggerNamingConvention.keyword

        def talk(self, *args, **kwargs):
            logger = self.logger
            logger.info(
                "I'm a class instance and my args are {} and my kwargs are {}".format(
                    args, kwargs
                )
            )
            logger.warning("logged class Test Warning")
            logger.error("logged class TEST ERROR")


def parsed_message_counts(df):
    parsed_msgs = df.shape[0]
    parsed_multiline = (df.message.str.count("\n") > 0).sum()
    parsed_exceptions = (~df.stackTrace.isnull()).sum()
    msgtypes = dict(sorted(Counter(df.levelname).items()))
    return parsed_msgs, parsed_multiline, parsed_exceptions, msgtypes


def all_message_counts():
    msg, multi, err, msgtypes = 0, 0, 0, Counter()
    for log in chain(_TestLogger.manager.loggerDict.values()):
        if hasattr(log, "name"):
            print(log, log.name, log.__class__)
        if isinstance(log, _TestLogger):
            msg += log.total
            multi += log.multiline
            err += log.stacktraces
            msgtypes += log.levelcounts

    return msg, multi, err, dict(sorted(msgtypes.items()))


def log_all_counts():
    for log in chain(_TestLogger.manager.loggerDict.values()):
        if isinstance(log, CountingLogger):
            log._report_stats()


def warn_logger_disabled(logger):
    if logger.disabled:
        warn("This logger is disabled!")
        warn("The logger and its handlers are {}, {}".format(logger, logger.handlers))
        warn(
            "The parent logger and its handlers are {}, {}".format(
                logger.parent, logger.parent.handlers
            )
        )


def test_logging_multiline1(logger):
    logger.debug("This is:\n\t\t\ta multiline log message!")


def test_logged_class():
    ti1 = get_test_instance("one")
    ti2 = get_test_instance("two")

    ti1.talk(1, 2, 3, foo="bar")
    ti2.talk(list(range(10)), bar="baz")

    ti1.talk("second arg is a logger", ti1.logger)


def test_logged_class_pickle(test_pickle_path):
    ti3 = get_test_instance("three")

    with open(test_pickle_path, "wb") as f:
        pickle.dump(ti3, f)

    with open(test_pickle_path, "rb") as f:
        ti4 = pickle.load(f)

    ti4.logger.info("Still kickin'!")
    assert ti4.__logname__ == ti3.__logname__
    assert ti4.logger.name == ti3.logger.name


def test_log_exception(logger):
    try:
        1 / 0
    except:
        logger.critical("You CANNOT do that!", exc_info=True)


def test_log_multiline(logger):
    logger.debug("Here's a multiline message -\n    see, this is the second line.")


def test_logfile_dataframe_parse(this_logfile, logger):
    logger.info("Now parsing {} to a dataframe".format(this_logfile))
    logger.debug(
        "parsing using regex:\n\t%s"
        % log_line_regex(DEFAULT_LOG_MSG_FMT, DEFAULT_LOG_DATE_FMT)
    )

    # this adds some METALOG-level messages to the file
    log_all_counts()
    # which should be parsed here
    parsed_log = log_file_to_df(
        this_logfile, datetime_index=True, raise_=False
    )

    assert_parsed_equals_logged(parsed_message_counts(parsed_log), all_message_counts())

    logger.info("parsed dataframe tail:\n{}".format(str(parsed_log.tail(2))))


def assert_parsed_equals_logged(parsed_message_counts, all_message_counts):
    total_msgs, total_multiline, total_exceptions, total_types = all_message_counts
    parsed_msgs, parsed_multiline, parsed_exceptions, parsed_types = (
        parsed_message_counts
    )
    assert total_multiline <= parsed_multiline
    assert total_exceptions <= parsed_exceptions
    assert total_msgs <= parsed_msgs
    for k, v in parsed_types.items():
        assert v >= total_types.get(k, 0)



# coding:utf-8
from .main import TypedIO, ArgSource, CLI, CONFIG, ENV
from .cli_nargs_ import cli_nargs, cli_option_nargs, cli_action
from .cli_complete import cli_completer
from .cli_repr_ import cli_repr

# import this last because it can register default implementations for the others
from .cli_parse import cli_parser
from .config_encode import config_encoder, config_key_encoder
from .config_decode import config_decoder, config_key_decoder
from .config_repr_ import config_repr
from .env_parse import env_parser
from .utils import File



# coding:utf-8
from logging import NOTSET, DEBUG, INFO, WARNING, ERROR, CRITICAL

# Useful default configurations

MAX_REPR_LEN = (
    24
)  # the max string len used for representing args in verbose function call
# signature logging

DEFAULT_PROGRESS_LOG_LEVEL = "INFO"

FULL_PATH_LOGGER_NAMING = (
    False
)  # use the fully qualified class path in the names of class/function
# loggers? Module name is already included in the default log format,
# which makes it an automatically accessible field in a log file parse

DEFAULT_LOG_DATE_FMT = "%y-%m-%d %H:%M:%S"

VERBOSE_LOG_MSG_FMT = "%(name)s:%(processName)s - %(filename)s:%(funcName)s:%(lineno)d - %(asctime)s,%(msecs)d - %(levelname)s:%(levelno)d - %(message)s"
DEFAULT_LOG_MSG_FMT = (
    "%(name)s - %(asctime)s,%(msecs)d - %(levelname)s:%(levelno)d - %(message)s"
)
DEFAULT_LOG_MSG_MP_FMT = "%(name)s:%(processName)s - %(asctime)s,%(msecs)d - %(levelname)s:%(levelno)d - %(message)s"


DEFAULT_ROOT_LOG_LEVEL = "INFO"
DEFAULT_CONSOLE_LOG_LEVEL = "INFO"
DEFAULT_FILE_LOG_LEVEL = "INFO"
DEFAULT_SMTP_LOG_LEVEL = "ERROR"
DEFAULT_SMTP_LOG_FLUSH_LEVEL = "CRITICAL"


METALOG_LEVEL = (
    "METALOG"
)  # level for CountingLogger to use when writing its own log message counts, set above
METALOG = NOTSET + (DEBUG - NOTSET) // 2
PROGRESS_LEVEL = "PROGRESS"  # level above info for reporting long-running job progress
PROGRESS = 25

TRACE_LEVEL = (
    "TRACE"
)  # for super fine-grained logging, e.g. of function signatures and return values
TRACE = 1

DEFAULT_FORMATTER_CONFIG = {
    "format": DEFAULT_LOG_MSG_FMT,
    "datefmt": DEFAULT_LOG_DATE_FMT,
}

DEFAULT_CONSOLE_HANDLER_CONFIG = {
    "class": "logging.StreamHandler",
    "formatter": "default",
    "level": DEFAULT_CONSOLE_LOG_LEVEL,
}

DEFAULT_CONSOLE_HANDLER_MP_CONFIG = {
    "class": "bourbaki.application.logging.handlers.MultiProcStreamHandler",
    "formatter": "default",
    "level": DEFAULT_CONSOLE_LOG_LEVEL,
}

# note that the file handler config needs a filename and optionally a maxBytes int
DEFAULT_FILE_HANDLER_CONFIG = {
    "class": "logging.handlers.RotatingFileHandler",
    "formatter": "default",
    "level": DEFAULT_FILE_LOG_LEVEL,
    "filename": None,
    "maxBytes": None,
}

DEFAULT_FILE_HANDLER_MP_CONFIG = {
    "class": "bourbaki.application.logging.handlers.MultiProcRotatingFileHandler",
    "formatter": "default",
    "level": DEFAULT_FILE_LOG_LEVEL,
    "filename": None,
    "maxBytes": None,
}

# note that the smtp handler config needs a capacity, mailhost, toaddrs
DEFAULT_SMTP_HANDLER_CONFIG = {
    "class": "bourbaki.application.logging.handlers.BufferingSMTPHandler",
    "formatter": "default",
    "level": DEFAULT_SMTP_LOG_LEVEL,
    "mailhost": None,
    "credentials": None,
    "toaddrs": None,
    "capacity": None,
    "secure": None,
    "flushLevel": DEFAULT_SMTP_LOG_FLUSH_LEVEL,
}

DEFAULT_SMTP_HANDLER_MP_CONFIG = {
    "class": "bourbaki.application.logging.handlers.MultiProcBufferingSMTPHandler",
    "formatter": "default",
    "level": DEFAULT_SMTP_LOG_LEVEL,
    "mailhost": None,
    "credentials": None,
    "toaddrs": None,
    "capacity": None,
    "secure": None,
    "flushLevel": DEFAULT_SMTP_LOG_FLUSH_LEVEL,
}

empty_config_dict = {
    "version": 1,
    "formatters": {},
    "handlers": {},
    "root": {"level": DEFAULT_CONSOLE_LOG_LEVEL},
}



# coding:utf-8
# Import commonly used things into the top-level namespace for quick import
from . import config, defaults, interface, timing, loggers, analysis
from .interface import Logged, InstanceLoggerNamingConvention
from .config import (
    configure_default_logging,
    configure_debug_logging,
    configure_custom_logging,
    enable_console_logging,
    disable_console_logging,
)
from .defaults import (
    DEFAULT_LOG_MSG_FMT,
    DEFAULT_LOG_DATE_FMT,
    DEFAULT_CONSOLE_LOG_LEVEL,
    DEFAULT_FILE_LOG_LEVEL,
)
from .timing import timed_context, TimedTaskContext
from .analysis import log_file_to_df
from .loggers import CountingLogger, ProgressLogger
from .handlers import (
    MemoryHandler,
    SMTPHandler,
    BufferingSMTPHandler,
    MultiProcHandler,
    MultiProcBufferingSMTPHandler,
    MultiProcStreamHandler,
    MultiProcRotatingFileHandler,
)



# coding:utf-8

from .io import load_config, dump_config, register_dump_config, register_load_config
from .io import (
    allow_unsafe_yaml,
    require_safe_yaml,
    ConfigFormat,
    LEGAL_CONFIG_EXTENSIONS,
)
from .python import (
    is_json_serializable,
    UnsafePythonSourceConfig,
    PythonConfigNotSerializable,
)
from .ini import INIConfigUnserializable



# coding:utf-8
from typing import Mapping, Iterable, Tuple, Any, Union
from collections.abc import Mapping as _Mapping
import argparse
from operator import itemgetter
from textwrap import indent
from multipledispatch import dispatch
from bourbaki.introspection.prettyprint import fmt_pyobj


class Namespace(Mapping[str, Any]):
    """
    Syntactic sugar for configuration objects; use [] or . to get field values.
    This also adds the option to include multiple config objects/dicts as positional args, whose keys override one
    another from left to right, and are finally overridden by any keyword args that are passed.
    """

    def __init__(
        self,
        *args: Union[Mapping, argparse.Namespace, Iterable[Tuple[str, Any]]],
        **kwargs: Any
    ):
        if not all(
            isinstance(d, (Mapping, argparse.Namespace, Iterable)) for d in args
        ):
            raise TypeError("all positional args must be dict or argparse.Namespace")

        for d in args:
            if isinstance(d, (Mapping, Iterable)):
                self.__dict__.update(d)
            elif isinstance(d, (Namespace, argparse.Namespace)):
                self.__dict__.update(d.__dict__)
            else:
                raise TypeError(
                    "varargs must each be instances of Mapping, argparse.Namespace, "
                    "or Iterable[Tuple[str, Any]]"
                )

        self.__dict__.update(kwargs)

    def __getitem__(self, item):
        return self.__dict__[item]

    def __setitem__(self, key, value):
        self.__dict__[key] = value

    def __delitem__(self, key):
        del self.__dict__[key]

    def get(self, item, default=None):
        return self.__dict__.get(item, default)

    def keys(self):
        return self.__dict__.keys()

    def values(self):
        return self.__dict__.values()

    def items(self):
        return self.__dict__.items()

    def __len__(self):
        return len(self.__dict__)

    def __contains__(self, item):
        return item in self.__dict__

    def __iter__(self):
        return iter(self.__dict__.keys())

    def __dir__(self):
        return list(self.__dict__.keys())

    def __repr__(self):
        return namespace_pretty_repr(self)


def namespace_pretty_repr(ns, indent_=4):
    name = ns.__class__.__name__
    break_ = "\n" if len(ns.__dict__) > 0 else ""
    return "{}({}{}{})".format(
        name,
        break_,
        indent(
            ",\n".join(
                "{}={}".format(
                    k,
                    namespace_pretty_repr(v, len(k) + 5)
                    if isinstance(v, Namespace)
                    else fmt_pyobj(v),
                )
                for k, v in sorted(type(ns).items(ns), key=itemgetter(0))
            ),
            " " * indent_,
        ),
        break_,
    )


@dispatch(_Mapping)
def namespace_recursive(obj: Mapping):
    return Namespace({k: namespace_recursive(v) for k, v in obj.items()})


@dispatch(argparse.Namespace)
def namespace_recursive(obj: argparse.Namespace):
    return namespace_recursive(obj.__dict__)


@dispatch((list, tuple))
def namespace_recursive(obj: list):
    return type(obj)(map(namespace_recursive, obj))


@dispatch(object)
def namespace_recursive(obj: object):
    return obj


@dispatch(_Mapping)
def un_namespace_recursive(obj: Mapping):
    return {k: un_namespace_recursive(v) for k, v in obj.items()}


@dispatch(argparse.Namespace)
def un_namespace_recursive(obj: argparse.Namespace):
    return un_namespace_recursive(obj.__dict__)


@dispatch((list, tuple))
def un_namespace_recursive(obj: list):
    return type(obj)(map(un_namespace_recursive, obj))


@dispatch(object)
def un_namespace_recursive(obj: object):
    return obj



# coding:utf-8
import os
import sys
from argparse import Action, SUPPRESS
from ..completion import install_shell_completion


# custom actions


def load_package_info(package):
    from importlib.metadata import distribution
    from pkginfo import Distribution

    meta = str(distribution(package).metadata)
    d = Distribution()
    d.parse(meta)
    return d.__dict__


def print_package_info(info, keys=None, exclude_keys=("package",)):
    if keys is None:
        value = {k: v for k, v in info.items() if k in info and k not in exclude_keys}
    elif isinstance(keys, str):
        value = info[keys]
    elif keys:
        value = {k: info[k] for k in keys if k in info and k not in exclude_keys}
    else:
        raise ValueError(
            "misunderstood keys arg; pass None, a string, or a collection of strings"
        )
    import yaml

    if isinstance(value, (int, str, bool, float, type(None))):
        print(value, file=sys.stdout)
    else:
        yaml.dump(value, sys.stdout, default_flow_style=False)


class InfoAction(Action):
    _info_type = "metadata"

    def __init__(
        self,
        option_strings,
        package,
        version=None,
        info_keys=None,
        dest=SUPPRESS,
        default=SUPPRESS,
        help=None,
    ):
        if help is None:
            help = "show {} for parent package {} and exit".format(
                self._info_type, repr(package)
            )

        super().__init__(
            option_strings=option_strings,
            dest=dest,
            default=default,
            nargs=0,
            help=help,
        )

        self.package = package
        self.version = version
        if version is True and info_keys and "version" not in info_keys:
            info_keys = ("version", *info_keys)
        self.info_keys = info_keys

    def load_package_info(self):
        info = load_package_info(self.package)
        if self.version is not None:
            if self.version is False:
                info.pop("version", None)
            elif self.version is not True:
                info["version"] = self.version
        return info

    def __call__(self, parser, namespace, values, option_string=None):
        info = self.load_package_info()
        print_package_info(info, keys=self.info_keys)
        sys.exit(0)


class PackageVersionAction(InfoAction):
    _info_type = "version"

    def __call__(self, *args, **kwargs):
        info = self.load_package_info()
        print_package_info(info, keys="version")
        sys.exit(0)


class InstallShellCompletionAction(Action):
    def __init__(
        self,
        option_strings,
        dest=SUPPRESS,
        default=SUPPRESS,
        help="Install bash completions for this CLI",
    ):
        super().__init__(
            option_strings=option_strings,
            dest=dest,
            default=default,
            nargs=0,
            help=help,
        )

    def __call__(self, parser, namespace, values, option_string=None):
        cmds = [parser.prog]
        src = getattr(parser, "source_file", None)
        if src is not None:
            cmds.append(os.path.basename(src))
        install_shell_completion(parser, *cmds)
        exit(0)


class SetExecuteFlagAction(Action):
    def __init__(
        self,
        option_strings,
        dest=SUPPRESS,
        default=SUPPRESS,
        help="only execute disk-altering, expensive, or external-resource-dependent commands if "
        "this flag is passed",
    ):
        super().__init__(
            option_strings=option_strings,
            dest=dest,
            default=default,
            nargs=0,
            help=help,
        )

    def __call__(self, parser, namespace, values, option_string=None):
        from bourbaki.application.cli import main

        main.EXECUTE = True
